import os
import importlib.util

def get_module_content(module_name):
    # 모듈 스펙 가져오기
    spec = importlib.util.find_spec(module_name)
    if spec and spec.origin:
        module_path = spec.origin
        # print(f"Module Path: {module_path}")
        
        # 파일 내용 읽기
        with open(module_path, 'r', encoding='utf-8') as file:
            content = file.read()
        return content

# 파일 경로 확인 및 내용 가져오기
origin_content = get_module_content(
    "aicastle.deepracer.vehicle.api.vehicle_control_module_content.origin"
)

custom_content = get_module_content(
    "aicastle.deepracer.vehicle.api.vehicle_control_module_content.custom"
)


"""
vehicle_control.py

This is the module that holds the APIs and utility functions required to capture
the action performed by user from the UI regarding the throttle, angle etc in the
manual drive and autonomous mode, and trigger appropriate service calls to move
the servo/motor.
"""

import math
from flask import (Blueprint,
                   jsonify,
                   request)

from deepracer_interfaces_pkg.msg import ServoCtrlMsg
from deepracer_interfaces_pkg.srv import (ActiveStateSrv,
                                          EnableStateSrv,
                                          NavThrottleSrv,
                                          GetCtrlModesSrv)
from webserver_pkg import constants
from webserver_pkg.utility import (api_fail,
                                   call_service_sync)
from webserver_pkg import webserver_publisher_node


VEHICLE_CONTROL_BLUEPRINT = Blueprint("vehicle_control", __name__)


def get_rescaled_manual_speed(categorized_throttle, max_speed_pct):
    """Return the non linearly rescaled speed value based on the max_speed_pct.

    Args:
        categorized_throttle (float): Float value ranging from -1.0 to 1.0.
        max_speed_pct (float): Float value ranging from 0.0 to 1.0 taken as input
                               from maximum speed input.
    Returns:
        float: Categorized value of the input speed.
    """
    """
    Example return values:
    categorized_throttle: -0.3 max_speed_pct: 0.1 mapped_speed: -0.1383742911153119
    categorized_throttle: -0.5 max_speed_pct: 0.1 mapped_speed: -0.224952741020794
    categorized_throttle: -0.8 max_speed_pct: 0.1 mapped_speed: -0.3463137996219282

    categorized_throttle: -0.3 max_speed_pct: 0.5 mapped_speed: -0.208
    categorized_throttle: -0.5 max_speed_pct: 0.5 mapped_speed: -0.33333333333333337
    categorized_throttle: -0.8 max_speed_pct: 0.5 mapped_speed: -0.5013333333333334

    categorized_throttle: -0.3 max_speed_pct: 1.0 mapped_speed: -0.552
    categorized_throttle: -0.5 max_speed_pct: 1.0 mapped_speed: -0.8
    categorized_throttle: -0.8 max_speed_pct: 1.0 mapped_speed: -0.992

    categorized_throttle: 0.3 max_speed_pct: 0.1 mapped_speed: 0.1383742911153119
    categorized_throttle: 0.5 max_speed_pct: 0.1 mapped_speed: 0.224952741020794
    categorized_throttle: 0.8 max_speed_pct: 0.1 mapped_speed: 0.3463137996219282

    categorized_throttle: 0.3 max_speed_pct: 0.5 mapped_speed: 0.208
    categorized_throttle: 0.5 max_speed_pct: 0.5 mapped_speed: 0.33333333333333337
    categorized_throttle: 0.8 max_speed_pct: 0.5 mapped_speed: 0.5013333333333334

    categorized_throttle: 0.3 max_speed_pct: 1.0 mapped_speed: 0.552
    categorized_throttle: 0.5 max_speed_pct: 1.0 mapped_speed: 0.8
    categorized_throttle: 0.8 max_speed_pct: 1.0 mapped_speed: 0.992
    """
    # return 0.0 if categorized_throttle or maximum speed pct is 0.0
    if categorized_throttle == 0.0 or max_speed_pct == 0.0:
        return 0.0

    # get the parameter value to calculate the coefficients a, b in the equation y=ax^2+bx
    # The lower the update_speed_scale_value parameter, higher the impact on the
    # final mapped_speed.
    # Hence the update_speed_scale_value parameter is inversely associated with max_speed_pct
    # and bounded by MANUAL_SPEED_SCALE_BOUNDS.
    # Ex: max_speed_pct = 0.5; update_speed_scale_value = 3
    #     max_speed_pct = 1.0; update_speed_scale_value = 1
    # Lower the update_speed_scale_value: categorized_throttle value gets mapped to
    # higher possible values.
    #   Example: update_speed_scale_value = 1.0;
    #            categorized_throttle = 0.8 ==> mapped_speed = 0.992
    # Higher the update_speed_scale_value: categorized_throttle value gets mapped to
    # lower possible values.
    #   Example: update_speed_scale_value = 3.0;
    #            categorized_throttle = 0.8 ==> mapped_speed = 0.501

    inverse_max_speed_pct = (1 - max_speed_pct)
    update_speed_scale_value = \
        constants.MANUAL_SPEED_SCALE_BOUNDS[0] + \
        inverse_max_speed_pct * \
        (constants.MANUAL_SPEED_SCALE_BOUNDS[1] - constants.MANUAL_SPEED_SCALE_BOUNDS[0])
    speed_mapping_coefficients = dict()

    # recreate the mapping coefficeints for the non-linear equation ax^2 + bx based on
    # the update_speed_scale_value.
    # These coefficents map the [update_speed_scale_value, update_speed_scale_value/2]
    # values to DEFAULT_SPEED_SCALE values [1.0, 0.8].
    speed_mapping_coefficients["a"] = \
        (1.0 / update_speed_scale_value**2) * \
        (2.0 * constants.DEFAULT_SPEED_SCALES[0] - 4.0 * constants.DEFAULT_SPEED_SCALES[1])
    speed_mapping_coefficients["b"] = \
        (1.0 / update_speed_scale_value) * \
        (4.0 * constants.DEFAULT_SPEED_SCALES[1] - constants.DEFAULT_SPEED_SCALES[0])
    return math.copysign(1.0, categorized_throttle) * \
        (speed_mapping_coefficients["a"] * abs(categorized_throttle)**2 +
         speed_mapping_coefficients["b"] * abs(categorized_throttle))


def get_categorized_manual_throttle(throttle):
    """Return the value of the category in which the input throttle belongs to.

    Args:
        input (float): Float value ranging from -1.0 to 1.0 taken as input from joystick.

    Returns:
        float: Value of the category the input throttle belonged to.
    """
    if abs(throttle) >= 0.8:
        throttle = math.copysign(0.8, throttle)
    elif abs(throttle) >= 0.5:
        throttle = math.copysign(0.5, throttle)
    elif abs(throttle) > 0.0:
        throttle = math.copysign(0.3, throttle)
    return throttle


def get_categorized_manual_angle(angle):
    """Return the value of the category in which the input angle belongs to.

    Args:
        input (float): Float value ranging from -1.0 to 1.0 taken as input from joystick.

    Returns:
        float: Value of the category the input angle belonged to.
    """
    if abs(angle) >= 0.8:
        angle = math.copysign(0.8, angle)
    elif abs(angle) >= 0.5:
        angle = math.copysign(0.5, angle)
    elif abs(angle) > 0:
        angle = math.copysign(0.3, angle)
    return angle


@VEHICLE_CONTROL_BLUEPRINT.route("/api/manual_drive", methods=["PUT", "POST"])
def api_manual_drive():
    """API that publishes control messages to control the angle and throttle in
       manual drive mode.

    Returns:
        dict: Execution status if the API call was successful.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    angle = request.json.get("angle")
    throttle = request.json.get("throttle")
    max_speed = request.json.get("max_speed")

    if angle is None:
        return api_fail("angle is required")
    if throttle is None:
        return api_fail("throttle is required")
    if max_speed is None:
        return api_fail("max_speed is required")

    if angle < -1.0 or angle > 1.0:
        return api_fail("angle out of range")
    if throttle < -1.0 or throttle > 1.0:
        return api_fail("throttle out of range")

    webserver_node.get_logger().info(f"Angle: {angle}  Throttle: {throttle}")

    # Create the servo message.
    msg = ServoCtrlMsg()
    # bound the throttle value based on the categories defined
    msg.angle = -1.0 * get_categorized_manual_angle(angle)
    categorized_throttle = get_categorized_manual_throttle(throttle)
    msg.throttle = -1.0 * get_rescaled_manual_speed(categorized_throttle, max_speed)
    webserver_node.pub_manual_drive.publish(msg)
    return jsonify({"success": True})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/drive_mode", methods=["PUT", "POST"])
def api_set_drive_mode():
    """API to toggle the drive mode between Autonomous/Manual mode.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    drive_mode = request.json.get("drive_mode")
    if drive_mode is None:
        return jsonify({"success": False, "reason": "drive_mode must be set."})

    webserver_node.get_logger().info(f"Changed the vehicle state to {drive_mode}")
    drive_mode_state = 0 if drive_mode == "manual" else 1

    try:
        vehicle_state_req = ActiveStateSrv.Request()
        vehicle_state_req.state = drive_mode_state
        vehicle_state_res = call_service_sync(webserver_node.vehicle_state_cli,
                                              vehicle_state_req)
        if vehicle_state_res and (vehicle_state_res.error == 0):
            return jsonify(success=True)
        else:
            webserver_node.get_logger().error("Vehicle state service call failed")
            return jsonify(success=False, reason="Error")

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach vehicle state server: {ex}")
        return jsonify({"success": False,
                        "reason": "Unable to reach vehicle state server."})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/start_stop", methods=["PUT", "POST"])
def api_set_start_stop():
    """API to call the enable_state service to start and stop the vehicle.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    start_stop = request.json.get("start_stop")
    if start_stop is None:
        return jsonify({"success": False, "reason": "start_stop must be set."})

    webserver_node.get_logger().info(f"Changed the enable state to {start_stop}")
    start_stop_state = False if start_stop == "stop" else True
    try:
        enable_state_req = EnableStateSrv.Request()
        enable_state_req.is_active = start_stop_state
        enable_state_res = call_service_sync(webserver_node.enable_state_cli, enable_state_req)
        if enable_state_res and (enable_state_res.error == 0):
            return jsonify({"success": True})
        else:
            return jsonify({"success": False, "reason": "Error"})

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach enable state server: {ex}")
        return jsonify({"success": False,
                        "reason": "Unable to reach enable state server."})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/max_nav_throttle", methods=["PUT", "POST"])
def max_nav_throttle():
    """API to call the navigation_throttle service to set the throttle scale in the
       autonomous mode.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    nav_throttle = request.json.get("throttle")
    if nav_throttle is None:
        return jsonify({"success": False, "reason": "value must be set."})
    webserver_node.get_logger().info(f"Setting max navigation throttle to {nav_throttle}")
    try:
        set_throttle_req = NavThrottleSrv.Request()
        set_throttle_req.throttle = nav_throttle / constants.MAX_AUTO_THROTTLE_VAL
        set_throttle_res = call_service_sync(webserver_node.set_throttle_cli, set_throttle_req)
        if set_throttle_res and (set_throttle_res.error == 0):
            return jsonify({"success": True})
        else:
            return jsonify(success=False, reason="Failed to call the navigation throttle service")
    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach navigation throttle server: {ex}")
        return jsonify(success=False, reason="Unable to reach navigation throttle server")


@VEHICLE_CONTROL_BLUEPRINT.route("/api/control_modes_available", methods=["GET"])
def control_modes_available():
    """API to call the GetCtrlModesCountSrv service to get the list of available modes
       in ctrl_pkg (autonomous/manual/calibration).

    Returns:
        dict: Execution status if the API call was successful, list of available modes
              and error reason if call fails.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    webserver_node.get_logger().info("Providing the number of available modes")
    try:
        get_ctrl_modes_req = GetCtrlModesSrv.Request()
        get_ctrl_modes_res = call_service_sync(webserver_node.get_ctrl_modes_cli,
                                               get_ctrl_modes_req)

        control_modes_available = list()
        for mode in get_ctrl_modes_res.modes:
            control_modes_available.append(constants.MODE_DICT[mode])

        data = {
            "control_modes_available": control_modes_available,
            "success": True
        }
        return jsonify(data)

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach get ctrl modes service: {ex}")
        return jsonify(success=False, reason="Error")



import logging
import sys
import requests
import urllib3
import json
import time
import cv2
import numpy as np
from requests_toolbelt.multipart.encoder import MultipartEncoder
from bs4 import BeautifulSoup
import os
import threading
from collections import deque
from pynput import keyboard
from datetime import datetime
import paramiko
import time
from tqdm import tqdm


from aicastle.deepracer.vehicle.api.controller import (
    DiscreteController,
    ContinuousController,
    KeyboardController
)
from aicastle.deepracer.vehicle.api.constant import (
    INFERENCE_IMG_HEIGHT,
    INFERENCE_IMG_WIDTH,
    MAX_IMG_HEIGHT,
    MAX_IMG_WIDTH,
    DEFAULT_IMG_HEIGHT,
    DEFAULT_IMG_WIDTH
)
from aicastle.deepracer.vehicle.api.ssh import (
    RebootTrigger,
    ssh_connect,
    ssh_sudo_command,
    ssh_remote_content_update,
    vehicle_control_module_content_update
)
from aicastle.deepracer.vehicle.api.utils import (
    KeyboardListener,
    get_time_now,
    show_image,
    show_info
)


logging.getLogger().addHandler(logging.StreamHandler(sys.stdout))
logging.getLogger().setLevel(logging.INFO)

class DeepracerVehicleApiError(Exception):
	pass


def get_csrf_token(session, URL):
    response = session.get(URL, verify=False, timeout=10)  # Cannot verify with Deep Racer
    # print(response.text)
    soup = BeautifulSoup(response.text, 'lxml')
    csrf_token = soup.select_one('meta[name="csrf-token"]')['content']
    return csrf_token

def get_headers(csrf_token, URL):
    headers = {
        'X-CSRFToken': csrf_token,
        "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36",
        "referer": URL + "/home",
        "origin": URL,
        "accept-encoding": "gzip, deflate, br",
        "content-type": "application/json;charset=UTF-8",
        "accept": "*/*",
        "sec-fetch-mode": "cors",
        "sec-fetch-site": "same-origin",
        "accept-language": "en-US,en;q=0.9",
        "x-requested-with": "XMLHttpRequest"
    }
    return headers



class VehicleClient:
    def __init__(self, 
            ip, 
            password, 
            speed_percent=50,
            speed_percent_increase=1,
            width=DEFAULT_IMG_WIDTH, 
            height=DEFAULT_IMG_HEIGHT, 
            stream_chunk_size=1024,
            frame_queue_maxlen=10,
            max_human_angle=30,
            max_human_speed=4,
            ssh_password='DeepRacer111!!!',
            ssh_module_update=False,
            ssh_module_update_custom=True,
        ):

        ########### 초기화 ###########
        self.ip = ip.strip()
        self.URL = "https://" + self.ip
        self.password = password
        self.ssh_password = ssh_password

        # step 1. 로그인
        self.login(retry=1, timeout=10)

        # step 2. vehicle_control_module 수정
        if ssh_module_update:
            original_isSshEnabled = self.is_ssh_enabled()
            self.enable_ssh()
            if not self.is_ssh_default_password_changed():
                self.reset_ssh_password(self.ssh_password)
            reboot_trigger = vehicle_control_module_content_update(
                hostname=self.ip, 
                username="deepracer", 
                password=self.ssh_password, 
                custom=ssh_module_update_custom
            )
            if original_isSshEnabled:
                self.enable_ssh()
            else :
                self.disable_ssh()
            if reboot_trigger:
                reboot_trigger.run()  # 재부팅
                self.login(retry=50, timeout=5)  # 재로그인

        ########### 설정 ###########
        self.speed_percent = speed_percent
        self.speed_percent_increase = speed_percent_increase
        self.width = width
        self.height = height
        self.stream_chunk_size = stream_chunk_size
        self.frame_queue_maxlen = frame_queue_maxlen
        self.max_human_angle = max_human_angle
        self.max_human_speed = max_human_speed

        # information
        self.info = {"ip": self.ip}

        # stream 데이터 가져오기 (실시간)
        self.get_stream()
        self.fetch_stream()
        
        # 배터리 레벨 모니터링 (실시간)
        self.get_battery_level()
        
        #  usb 연결 상태 모니터링 (실시간)
        self.get_usb_connected()

        # 센서 상태 모니터링 (실시간)
        self.get_sensor_status()

        # drive 초기화
        self.model_loading_status = None
        self.model = None
        self.drive_mode = None
        self.last_move_kwargs = {"angle":0, "speed":0}
        self.set_speed_percent(self.speed_percent)
        self.get_calibration_angle()
        self.get_calibration_speed()

        #     "ip": self.ip,
        #     "speed_percent": self.speed_percent,
        #     "drive_mode": self.drive_mode,
        #     "model": self.model,
        #     "fetch_fps": self.fetch_fps,
        #     "battery_level": self.battery_level,
        #     "usb_connected": self.usb_connected,
        #     "camera_status": self.camera_status,
        #     "lidar_status": self.lidar_status,
        #     "stereo_status": self.stereo_status,
        #     "angle_center": self.angle_center,
        #     "angle_left": self.angle_left,
        #     "angle_right": self.angle_right,
        #     "speed_stopped": self.speed_stopped,
        #     "speed_forward": self.speed_forward,
        #     "speed_backward": self.speed_backward
        # }

    def login(self, retry=1, timeout=10):
        print(f"Login to {self.ip}", end=" ")
        login_success = False
        for i in range(retry):
            print(".", end="")
            try :
                self.session = requests.Session()
                urllib3.disable_warnings()
                self.csrf_token = get_csrf_token(self.session, self.URL)
                self.headers = get_headers(self.csrf_token, self.URL)
                login_post = self.session.post(
                    url= self.URL + "/login", 
                    data={'password': self.password}, 
                    # headers = self.headers,
                    headers={
                        'X-CSRFToken': self.csrf_token,
                        "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/76.0.3809.100 Safari/537.36"
                    }, 
                    verify=False,
                    timeout=timeout
                )
                if login_post.status_code != 200:
                    raise DeepracerVehicleApiError("Log in failed. Error message {}".format(login_post.text))
                else :
                    login_success = True
                    break
            except :
                time.sleep(timeout)

        if login_success:
            print("Success!")
            return True
        else :
            raise DeepracerVehicleApiError("Log in failed")

    ############# ssh #############
    def is_ssh_enabled(self, timeout=20):
        response = self._get(path="/api/isSshEnabled", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return response['isSshEnabled']
        else :
            raise DeepracerVehicleApiError("Failed to get ssh status")
    
    def enable_ssh(self, timeout=20):
        response = self._get(path="/api/enableSsh", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return response
        else :
            raise DeepracerVehicleApiError("Failed to enable ssh")

    def disable_ssh(self, timeout=20):
        response = self._get(path="/api/disableSsh", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return response['isSshEnabled']
        else :
            raise DeepracerVehicleApiError("Failed to disable ssh")

    def is_ssh_default_password_changed(self, timeout=20):
        response = self._get(path="/api/isSshDefaultPasswordChanged", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return response['isDefaultSshPasswordChanged']
        else :
            raise DeepracerVehicleApiError("Failed to get ssh default password status")
        
    def reset_ssh_password(self, ssh_password=None, timeout=30):
        ssh_password = self.ssh_password if ssh_password is None else ssh_password
        if self.is_ssh_default_password_changed():
            return True

        data = {"oldPassword":'deepracer', "newPassword":ssh_password}
        response = self._put(path="/api/resetSshPassword", headers=self.headers, data=data, timeout=timeout, raise_exception=True)
        if response and response.json()['success']:
            return True
        else :
            raise DeepracerVehicleApiError("Failed to reset ssh password")

    ############# stream #############
    def get_stream(self):
        image_url = self.URL + f"/route?topic=/camera_pkg/display_mjpeg&width={self.width}&height={self.height}"
        # lidar_overlay_url = f"/route?topic=/sensor_fusion_pkg/overlay_msg&width={self.width}&height={self.height}"
        # object_detection_url = f"/route?topic=/object_detection_pkg/detection_display&width={self.width}&height={self.height}"
        # qr_detection_url = f"/route?topic=/qr_detection_pkg/qr_detection_display&width={self.width}&height={self.height}"
        self.stream = self.session.get(image_url, headers=self.headers, stream=True, verify=False)
        if not self.stream.status_code == 200:
            raise Exception()

    def fetch_stream(self, async_=True, waiting_time=10):
        if async_:
            self.frame_queue = deque(maxlen=self.frame_queue_maxlen)
            self.fetch_stop_event = threading.Event()
            self.fetch_fps = None
            self.fetch_stream_thread = threading.Thread(
                target=self.fetch_stream,
                kwargs={"async_":False},
                daemon=True
            )
            self.fetch_stream_thread.start()
            s_time = time.time()
            while True:
                if len(self.frame_queue) > 0 :
                    return
                elif (time.time() - s_time) > waiting_time :
                    raise Exception()
                else :
                    time.sleep(0.1)
        else :
            byte_data = bytes()
            self.spf_deque = deque(maxlen=30)
            s_time = time.time()
            for chunk in self.stream.iter_content(chunk_size=self.stream_chunk_size):
                byte_data += chunk
                a = byte_data.find(b'\xff\xd8')
                b = byte_data.find(b'\xff\xd9')
                if a != -1 and b != -1:
                    jpg = byte_data[a:b+2]
                    byte_data = byte_data[b+2:]
                    img = cv2.imdecode(np.frombuffer(jpg, dtype=np.uint8), cv2.IMREAD_COLOR)
                    self.frame_queue.append(img)
                    f_time = time.time()
                    self.spf_deque.append(f_time-s_time)
                    fetch_spf = max(sum(self.spf_deque) / len(self.spf_deque), 0.01)
                    self.fetch_fps = round(1 / fetch_spf)
                    self.info["fetch_fps"] = self.fetch_fps
                    s_time = time.time()

                    # 평균값 내고 fps 계산
                if self.fetch_stop_event.is_set():
                    break


    ############# image #############
    def get_image(self, width=None, height=None, color='bgr'):
        width = width if width else self.width
        height = height if height else self.height
        img = self.frame_queue[-1]
        if (width != self.width) or (height != self.height):
            img = cv2.resize(img, (width, height))
        
        if color == 'bgr':
            return img
        elif color == 'rgb':
            return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        elif color == 'gray':
            return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    def save_image(self, save_dir=None, format='jpg', width=None, height=None):
        if save_dir is None:
            save_dir = "data/images/"
            os.makedirs(save_dir, exist_ok=True)
        save_time = get_time_now(milliseconds=True)
        path = os.path.join(save_dir, f"{save_time}.{format}")
        img = self.get_image(width=width, height=height, color='bgr')
        cv2.imwrite(path, img)
        return path

    ############# request #############
    def _get(self, path, headers=None, timeout=10, raise_exception=False):
        url = self.URL + path
        logging.debug("> Get %s",url)
        headers = headers if headers is None else self.headers
        try :
            response = self.session.get(url=url, headers=headers, verify=False, timeout=timeout)
            return response
        except Exception as e:
            if raise_exception:
                raise e
            return None
        
    def _put(self, path, headers=None, data=None, json_content=True, timeout=10, raise_exception=False):
        url = self.URL + path
        logging.debug("> Put %s with %s",url, data)
        headers = self.headers if headers is None else headers
        data = {} if data is None else data
        try :
            if json_content:
                response = self.session.put(url, json=data, headers=headers, verify=False, timeout=timeout)
            else :
                response = self.session.put(url, data=data, headers=headers, verify=False, timeout=timeout)
            return response
        except Exception as e:
            if raise_exception:
                raise e
            return None

    ############# control #############
    def stop(self, timeout=5):
        response = self._put(path="/api/start_stop", data={'start_stop': "stop"}, timeout=timeout)
        if self.drive_mode == "manual":
            self.drive_mode = None
            self.last_move_kwargs = {"angle":0, "speed":0}
        return response

    # autonomous mode
    def start(self, timeout=5):
        response = self._put(path="/api/start_stop", data={'start_stop': "start"}, timeout=timeout)
        return response

    def get_speed_percent(self):
        return self.speed_percent

    def set_speed_percent(self, speed_percent, timeout=1):
        self.speed_percent = max(min(speed_percent, 100), 1)
        self.info["speed_percent"] = self.speed_percent
        # if self.drive_mode == "auto":
        data = {"throttle":self.speed_percent}
        self._put(path="/api/max_nav_throttle", data = data, timeout=timeout)
        if self.drive_mode == "manual":
            self.move(**self.last_move_kwargs)
        return self.speed_percent
    
    def speed_percent_up(self, speed_percent_increase=None):
        speed_percent_increase = speed_percent_increase if speed_percent_increase is not None else self.speed_percent_increase
        self.set_speed_percent(self.speed_percent + speed_percent_increase)
        return self.speed_percent
    
    def speed_percent_down(self, speed_percent_increase=None):
        speed_percent_increase = speed_percent_increase if speed_percent_increase is not None else self.speed_percent_increase
        self.set_speed_percent(self.speed_percent - speed_percent_increase)
        return self.speed_percent
    
    ############# calibration #############
    def calibration_mode(self, timeout=5):
        self.stop()
        response = self._get(path="/api/set_calibration_mode", timeout=timeout)
        self.drive_mode = "calibration"
        self.info["drive_mode"] = self.drive_mode
        return response

    def get_calibration_angle(self, timeout=5):
        if self.drive_mode != "calibration":
            self.calibration_mode()
        response = self._get(path="/api/get_calibration/angle", timeout=timeout)
        response_data = response.json()
        self.angle_center = response_data['mid']
        self.angle_left = response_data['max']
        self.angle_right = response_data['min']
        self.info["angle_center"] = self.angle_center
        self.info["angle_left"] = self.angle_left
        self.info["angle_right"] = self.angle_right
        return {"center":self.angle_center, "left":self.angle_left, "right":self.angle_right}

    def get_calibration_speed(self, timeout=5):
        if self.drive_mode != "calibration":
            self.calibration_mode()
        response = self._get(path="/api/get_calibration/throttle", timeout=timeout)
        response_data = response.json()
        self.speed_stopped = response_data['mid']
        self.speed_forward = response_data['min']
        self.speed_backward = response_data['max']
        self.info["speed_stopped"] = self.speed_stopped
        self.info["speed_forward"] = self.speed_forward
        self.info["speed_backward"] = self.speed_backward
        return {"stopped":self.speed_stopped, "forward":self.speed_forward, "backward":self.speed_backward}
    
    def set_calibration_angle(self, center=None, left=None, right=None, polarity=1, timeout=5):
        self.get_calibration_angle()
        data = {
            "max": left if left is not None else self.angle_left,
            "mid": center if center is not None else self.angle_center,
            "min": right if right is not None else self.angle_right,
            "polarity": polarity
        }
        self._put(path="/api/set_calibration/angle", data=data, timeout=timeout)
        return self.get_calibration_angle()
    
    def set_calibration_speed(self, stopped=None, forward=None, backward=None, polarity=-1, timeout=5):
        self.get_calibration_speed()
        data = {
            "max": backward if backward is not None else self.speed_backward,
            "mid": stopped if stopped is not None else self.speed_stopped,
            "min": forward if forward is not None else self.speed_forward,
            "polarity": polarity
        }
        self._put(path="/api/set_calibration/throttle", data=data, timeout=timeout)
        return self.get_calibration_speed()

    ##### auto mode
    def auto_mode(self, timeout=1):
        self.stop()
        response = self._put(path="/api/drive_mode",data={'drive_mode': "auto"}, timeout=timeout)
        self.drive_mode = "auto"
        self.info["drive_mode"] = self.drive_mode
        self.set_speed_percent(self.speed_percent)
        return response
                
    ##### manual mode
    def manual_mode(self, timeout=2):
        self.stop()
        response = self._put(path="/api/drive_mode",data={'drive_mode': "manual"}, timeout=timeout)
        self.drive_mode = "manual"
        self.last_move_kwargs = {"angle":0, "speed":0}
        self.info["drive_mode"] = self.drive_mode
        self.start()
        return response
    
    def move(self, angle, speed, scale='human', timeout=1):
        if self.drive_mode != "manual":
            self.manual_mode()

        self.last_move_kwargs = {"angle":angle, "speed":speed, "scale":scale, "timeout":timeout}
        if scale == 'human':
            angle = angle / self.max_human_angle
            speed = speed / self.max_human_speed
        elif scale == 'robot':
            angle = angle
            speed = speed
        else:
            raise DeepracerVehicleApiError("Invalid scale")
        data = {'angle': -max(-1, min(1, angle)), 'throttle': -max(-1, min(1, speed)), "max_speed":self.speed_percent / 100}
        response = self._put(path="/api/manual_drive", data = data, timeout=timeout)
        return response

    def move_forward(self, duration=1):
        self.move(0, 0.7, scale='robot')
        time.sleep(duration)
        self.move(0, 0, scale='robot')

    def move_backward(self, duration=1):
        self.move(0, -0.7, scale='robot')
        time.sleep(duration)
        self.move(0, 0, scale='robot')

    def move_left(self, duration=1):
        self.move(1, 0.7, scale='robot')
        time.sleep(duration)
        self.move(0, 0, scale='robot')

    def move_right(self, duration=1):        
        self.move(-1, 0.7, scale='robot')
        time.sleep(duration)
        self.move(0, 0, scale='robot')

    ############# model #############
    def get_model_list(self, timeout=5):
        response = self._get(path="/api/models", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return response
        else :
            return None

    def upload_model(self, model_zip_path):
        model_file = open(model_zip_path, 'rb')
        model_basename = os.path.basename(model_zip_path)
        model_folder_name = model_basename.split('.')[0]
        headers = self.headers.copy()
        multipart_data = MultipartEncoder(
            fields={
                # a file upload field
                'file': (model_basename, model_file, None)
            }
        )
        headers['content-type'] = multipart_data.content_type
        # response = self.session.put(self.URL + "/api/uploadModels", data=multipart_data, headers=headers, verify=False, timeout=60)
        response = self._put(path="/api/uploadModels", headers=headers, data=multipart_data, json_content=False, timeout=60)
        if response and response.json()['success']:
            return model_folder_name
        else :
            raise DeepracerVehicleApiError("Failed to upload model")
    
    def uploaded_model_list(self, timeout=5):
        response = self._get(path="/api/uploaded_model_list", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            return {model_info["name"]:model_info for model_info in response}
        else :
            raise DeepracerVehicleApiError("Failed to get uploaded model list")

    def is_model_loading(self, timeout=5):
        response = self._get(path="/api/isModelLoading", timeout=timeout)
        if response and response.status_code == 200:
            response = response.json()
            model_loading_status = response['isModelLoading']
            self.model_loading_status = model_loading_status
            return model_loading_status
        else :
            raise DeepracerVehicleApiError("Failed to get model loading status")

    def load_model(self, model_folder_name, model_name='model'):
        # auto mode로 변경
        if self.drive_mode != 'auto':
            response = self.auto_mode(timeout=10)
            if not response or response.status_code != 200:
                raise DeepracerVehicleApiError("Failed to set auto mode")
        else :
            self.stop()
        
        # 이미 로드된 모델인지 확인인
        if self.model_loading_status == 'loaded':
            if self.model == model_folder_name:
                return

        # 모델 로드
        response = self._put(path=f"/api/models/{model_folder_name}/{model_name}", timeout=10)
        if response and response.status_code == 200:
            # 모델 로드 확인
            loaded = False
            for i in range(100):
                try :
                    model_loading_status = self.is_model_loading()
                except :
                    time.sleep(0.5)
                    continue
                
                if model_loading_status == 'loaded':
                    loaded = True
                    self.model = model_folder_name
                    self.info["model"] = self.model
                    break  
                elif model_loading_status == 'loading':
                    time.sleep(0.5)
                elif model_loading_status == 'error':
                    raise DeepracerVehicleApiError("Failed to load model")
            if not loaded:
                raise DeepracerVehicleApiError("Failed to load model")
        else:
            raise DeepracerVehicleApiError("Failed to load model")


    ############ deviece info ############
    def get_battery_level(self, async_=True, timeout=5):
        if async_ :
            self.battery_level = None
            thread = threading.Thread(
                target=self.get_battery_level,
                kwargs={"async_":False, "timeout":timeout}
            )
            thread.start()
        else :
            while True :
                response = self._get(path="/api/get_battery_level", timeout=timeout)
                if response and response.status_code == 200:
                    response = response.json()
                    self.battery_level = response['battery_level']
                else :
                    self.battery_level = None

                self.info["battery_level"] = self.battery_level 
                time.sleep(timeout)

    def get_usb_connected(self, async_=True, timeout=5):
        if async_ :
            self.usb_connected = None
            thread = threading.Thread(
                target=self.get_usb_connected,
                kwargs={"async_":False, "timeout":timeout}
            )
            thread.start()
        else :
            while True :
                response = self._get(path="/api/is_usb_connected", timeout=timeout)
                if response and response.status_code == 200:
                    response = response.json()
                    self.usb_connected = response['is_usb_connected']
                else :
                    self.usb_connected = None

                self.info["usb_connected"] = self.usb_connected
                time.sleep(timeout)

    def get_sensor_status(self, async_=True, timeout=5):
        if async_ :
            self.camera_status = None
            self.lidar_status = None
            self.stereo_status = None
            thread = threading.Thread(
                target=self.get_sensor_status,
                kwargs={"async_":False, "timeout":timeout}
            )
            thread.start()
        else :
            while True :
                response = self._get(path="/api/get_sensor_status", timeout=timeout)
                if response and response.status_code == 200:
                    response = response.json()
                    self.camera_status = response['camera_status']
                    self.lidar_status = response['lidar_status']
                    self.stereo_status = response['stereo_status']
                else :
                    self.camera_status = None
                    self.lidar_status = None
                    self.stereo_status = None

                self.info["camera_status"] = self.camera_status
                self.info["lidar_status"] = self.lidar_status
                self.info["stereo_status"] = self.stereo_status
                time.sleep(timeout)


import cv2
import sys
import time
import threading
from datetime import datetime
from pynput import keyboard
import os
import tarfile
import base64
import math

class KeyboardListener: 
    def __init__(self, on_press_callback=None, on_release_callback=None):
        self.on_press_callback = on_press_callback if on_press_callback is not None else {}
        self.on_release_callback = on_release_callback if on_release_callback is not None else {}
        self.pressed_key = None
        self.pressed_keys = set()
        self.info = {"pressed_key": self.pressed_key, "pressed_keys": self.pressed_keys}
        self.listener = keyboard.Listener(on_press=self.on_press, on_release=self.on_release)
        self.listener.start()

    def on_press(self, key):
        try:
            key_str = key.char  # 일반 키
        except :
            key_str = str(key)  # 특수 키
        self.pressed_key = key_str
        self.pressed_keys.add(key_str)
        if key_str in self.on_press_callback:
            for target_function in self.on_press_callback[key_str]:
                target_function()
        self.info["pressed_key"] = self.pressed_key
        self.info["pressed_keys"] = self.pressed_keys

    def on_release(self, key):
        try:
            key_str = key.char  # 일반 키
        except :
            key_str = str(key)  # 특수 키
        self.pressed_key = None if key_str == self.pressed_key else self.pressed_key
        self.pressed_keys.discard(key_str)
        if key_str in self.on_release_callback:
            for target_function in self.on_release_callback[key_str]:
                target_function()
        self.info["pressed_key"] = self.pressed_key
        self.info["pressed_keys"] = self.pressed_keys


def get_time_now(milliseconds=True):
    if milliseconds:
        return datetime.now().strftime('%Y-%m-%dT%H-%M-%S-%f')[:-3]
    else:
        return datetime.now().strftime('%Y-%m-%dT%H-%M-%S')


def show_image(frame_queue, window_name="show_image", stop_key=None, stop_event=None, async_=True):
    if async_ :
        stop_event = threading.Event()
        thread = threading.Thread(
            target = show_image,
            kwargs={"frame_queue":frame_queue, "window_name":window_name, "stop_key":stop_key, "stop_event":stop_event, "async_":False}
        )
        thread.start()
        return stop_event
    else :
        # 창 이름 정의 및 크기 조정 가능 설정
        cv2.namedWindow(window_name, cv2.WINDOW_NORMAL)  # 창 크기 조정 가능
        cv2.setWindowProperty(window_name, cv2.WND_PROP_ASPECT_RATIO, cv2.WINDOW_KEEPRATIO)
        while True:
            if len(frame_queue) > 0:
                img = frame_queue[-1]
                cv2.imshow(window_name, img)
                key = cv2.waitKey(1)
                if cv2.getWindowProperty(window_name, cv2.WND_PROP_VISIBLE) < 1 or (stop_key is not None and key == stop_key) or stop_event.is_set():
                    break
            else :
                time.sleep(0.1)

        # cv2.destroyWindow(window_name)


def clear_output():
    if 'ipykernel' in sys.modules:
        # Jupyter Notebook 환경
        from IPython.display import clear_output as jupyter_clear_output
        jupyter_clear_output(wait=True)
    else:
        # 일반 터미널 환경
        os.system('cls' if os.name == 'nt' else 'clear')

def show_info(info_containers, show_keys=None, print_interval=0.1, ljust=10, stop_event=None, async_=True):
    if async_:
        if show_keys is None:
            show_keys = []
            for container in info_containers:
                show_keys += list(container.keys())
        stop_event = threading.Event()
        thread = threading.Thread(
            target=show_info,
            kwargs={"info_containers": info_containers, "show_keys": show_keys, "print_interval": print_interval, "ljust":ljust, "stop_event": stop_event, "async_": False},
        )
        thread.start()
        return stop_event
    
    else :
        while not stop_event.is_set():
            container_combine = {}
            for container in info_containers:
                for key in show_keys:
                    if key in container:
                        container_combine[key] = container[key]

            output = "\r|"
            for key in container_combine:
                output += f"{key}: {str(container_combine[key]).ljust(ljust)} | "
            sys.stdout.write(output)
            sys.stdout.flush()
            time.sleep(print_interval)

        sys.stdout.write(f"\r{'show info finished.....'.ljust(100)}") 
        sys.stdout.flush()

def extract_model(model_zip_path, model_folder_path=None):
    """tar.gz 모델 파일을 지정된 폴더로 추출."""
    model_folder_name = os.path.basename(model_zip_path).split('.')[0]
    if model_folder_path is None:
        model_folder_path = os.path.join(os.path.dirname(model_zip_path), model_folder_name)
    os.makedirs(model_folder_path, exist_ok=True)
    with tarfile.open(model_zip_path, 'r:gz') as tar:
        tar.extractall(model_folder_path)
    return model_folder_path


def read_image(image_path, width=None, height=None, color="rgb"):
    image = cv2.imread(image_path)

    # resize
    h, w = image.shape[:2]
    if width and height:
        image = cv2.resize(image, (width, height))
    elif width:
        ratio = width / w
        image = cv2.resize(image, (width, math.ceil(h * ratio)))
    elif height:
        ratio = height / h
        image = cv2.resize(image, (math.ceil(w * ratio), height))
    
    # color
    if color.lower() == "rgb":
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    elif color.lower() == "gray":
        image = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)   
    
    return image


def get_image_base64(image, format=".jpg"):
    image_encoded = cv2.imencode(f".{format}", image)[1]
    image_base64 = base64.b64encode(image_encoded).decode("utf-8")
    return image_base64




import json
import time
import os
import time
from aicastle.deepracer.vehicle.api.constant import (
    INFERENCE_IMG_HEIGHT,
    INFERENCE_IMG_WIDTH,
    MAX_IMG_HEIGHT,
    MAX_IMG_WIDTH,
    DEFAULT_IMG_HEIGHT,
    DEFAULT_IMG_WIDTH
)
from aicastle.deepracer.vehicle.api.utils import (
    KeyboardListener,
    get_time_now,
    show_image,
    show_info
)

class ManualController:
    def __init__(self,
            vehicle,
            move_interval=0.05,
            move_timeout=1,
            action_space=None,
            action_space_type=None, # continuous, discrete
            save=False,
            save_dir='data/images/',
            save_with_time_dir=True, # save_dir에 시간을 추가하여 디렉토리 생성
            save_format='jpg',
            save_interval=0.5, # 저장 간격 (초)
            save_width=None,
            save_height=None,
        ):
        self.vehicle = vehicle
        self.move_interval = move_interval
        self.move_timeout = move_timeout
        self.action_space = action_space
        self.action_space_type = action_space_type
        if self.action_space_type == 'discrete':
            self.index_action_map = {action["index"]: action for action in self.action_space}
        self.save = save
        self.save_dir = save_dir
        self.save_with_time_dir = save_with_time_dir
        self.save_format = save_format
        self.save_interval = save_interval
        self.save_width = save_width
        self.save_height = save_height
        if self.save:
            self.save_setting()
            self.save_setting_done = True
        else :
            self.save_setting_done = False

        self.angle = 0
        self.speed = 0
        self.last_move_time = -1
        self.last_save_time = -1

    def save_setting(self):
        if self.save_with_time_dir : 
            self.save_dir = os.path.join(self.save_dir, get_time_now(milliseconds=False))
        os.makedirs(self.save_dir, exist_ok=True)
        print(f"save_dir: {self.save_dir}")
        # metadata.json 저장
        with open(os.path.join(self.save_dir, "metadata.json"), 'w') as f:
            metadata = {
                "action_space": self.action_space,   
                "action_space_type": self.action_space_type,
                "save_format": self.save_format,
                "save_width": self.save_width,
                "save_height": self.save_height,
                "save_interval": self.save_interval
            }
            json.dump(metadata, f, indent=2)

        # 폴더 세팅
        if self.action_space_type == 'continuous':
            os.makedirs(os.path.join(self.save_dir, "images"), exist_ok=True)
            os.makedirs(os.path.join(self.save_dir, "labels"), exist_ok=True)
        
        if self.action_space_type == 'discrete':
            for action in self.action_space:
                os.makedirs(os.path.join(self.save_dir, str(action["index"])), exist_ok=True)

    def move_continuous(self, angle=None, speed=None, move_timeout=None, save=None):
        if (angle is None) and (speed is None):
            raise Exception("At least one of angle, speed must be entered.")
        self.angle = self.angle if angle is None else angle
        self.speed = self.speed if speed is None else speed
        move_timeout = self.move_timeout if move_timeout is None else move_timeout
        last_time = time.time()
        
        ##### move
        if (last_time - self.last_move_time) > self.move_interval:
            self.vehicle.move(
                angle = self.angle, 
                speed = self.speed,
                scale = 'human', 
                timeout = move_timeout,
            )
            self.last_move_time = last_time

        ##### save
        save = self.save if save is None else save
        if save and (not self.save_setting_done):
            self.save_setting()
            self.save_setting_done = True
        if save and (last_time - self.last_save_time) > self.save_interval:
            image_save_dir = os.path.join(self.save_dir, "images")
            try :
                image_save_path = self.vehicle.save_image(
                    save_dir=image_save_dir, 
                    format=self.save_format, 
                    width=self.save_width, 
                    height=self.save_height
                )
            except :
                image_save_path = None

            if image_save_path:
                label_save_path = os.path.join(self.save_dir, "labels", f"{os.path.basename(image_save_path).split('.')[0]}.txt")
                with open(label_save_path, 'w') as f:
                    f.write(f"{self.angle} {self.speed}")
                self.last_save_time = last_time

    def move_discrete(self, action_index=None, move_timeout=None, save=None):
        if action_index is None:
            raise Exception("action_index must be entered.")
        action = self.index_action_map[action_index]
        angle = action["steering_angle"]
        speed = action["speed"]
        self.angle = angle
        self.speed = speed
        move_timeout = self.move_timeout if move_timeout is None else move_timeout
        last_time = time.time()
        
        ##### move
        if (last_time - self.last_move_time) > self.move_interval:
            self.vehicle.move(
                angle = self.angle, 
                speed = self.speed,
                scale = 'human', 
                timeout = move_timeout,
            )
            self.last_move_time = last_time

        ##### save
        save = self.save if save is None else save
        if save and (not self.save_setting_done):
            self.save_setting()
            self.save_setting_done = True
        if save and (last_time - self.last_save_time) > self.save_interval:
            if action_index in self.index_action_map:
                image_save_dir = os.path.join(self.save_dir, str(action_index))
                try :
                    image_save_path = self.vehicle.save_image(
                        save_dir=image_save_dir, 
                        format=self.save_format, 
                        width=self.save_width, 
                        height=self.save_height
                    )
                except :
                    image_save_path = None

                if image_save_path: 
                    self.last_save_time = last_time

    def stop(self):
        self.vehicle.move(angle=0, speed=0, scale='human', timeout=self.move_timeout)


class DiscreteController(ManualController):
    def __init__(self, **kwargs):
        kwargs["action_space_type"] = 'discrete'
        kwargs["action_space"] = [
            {"steering_angle": 0, "speed": 2, "index": 0},
            {"steering_angle": 30, "speed": 2, "index": 1},
            {"steering_angle": -30, "speed": 2, "index": 2},
            {"steering_angle": 0, "speed": -2, "index": 3},
        ] if kwargs["action_space"] is None else kwargs["action_space"]
        super().__init__(**kwargs)

    def move(self, action_index, **kwargs):
        super().move_discrete(action_index=action_index, **kwargs)


class ContinuousController(ManualController):
    def __init__(self, **kwargs):
        kwargs["action_space_type"] = 'continuous'
        kwargs["action_space"] = {
            "steering_angle": {
                "high": 30, 
                "low": -30
            },
            "speed": {
                "high": 4, 
                "low": 0.5
            },
        }
        super().__init__(**kwargs)

    def move(self, angle=None, speed=None, **kwargs):
        angle = min(self.action_space["steering_angle"]["high"], max(self.action_space["steering_angle"]["low"], angle))
        speed = min(self.action_space["speed"]["high"], max(self.action_space["speed"]["low"], speed))
        super().move_continuous(angle=angle, speed=speed, **kwargs)


class KeyboardController(DiscreteController):
    def __init__(self, keyboard_listener, index_key_map=None, **kwargs):
        self.keyboard_listener = keyboard_listener
        super().__init__(**kwargs)

        self.index_key_map = {
            0: "Key.up",
            1: "Key.left",
            2: "Key.right",
            3: "Key.down",
        } if index_key_map is None else index_key_map
        self.key_action_map = {self.index_key_map[action["index"]]: action for action in self.action_space}

    def run(self, exit_key='Key.esc', move_timeout=None, save=None):
        print(f"Press {exit_key} to stop")
        running = False
        while True:
            pressed_key = self.keyboard_listener.pressed_key
            if pressed_key == exit_key:
                break
            
            if pressed_key and pressed_key in self.key_action_map:
                action_index = self.key_action_map[pressed_key]["index"]
                response = self.move(
                    action_index = action_index, 
                    move_timeout=move_timeout, 
                    save=save,
                )
                running = True
            else:
                if running:
                    response = self.stop()
                    if response and response.status_code == 200:
                        running = False


class AutoModelLoadController:
    def __init__(self, vehicle, key_model_map, keyboard_listener):
        self.vehicle = vehicle
        self.key_model_map = key_model_map
        self.keyboard_listener = keyboard_listener
        self.loaded_model = None

    def run(self, exit_key='Key.esc'):
        print(f"Press {exit_key} to exit")
        while True:
            pressed_key = self.keyboard_listener.pressed_key
            if pressed_key == exit_key:
                break
            elif pressed_key and pressed_key in self.key_model_map:
                model_folder_name = self.key_model_map[pressed_key]
                if self.loaded_model != model_folder_name:
                    print(f"Model Loading: {model_folder_name}")
                    self.vehicle.load_model(model_folder_name)
                    self.loaded_model = model_folder_name
                    print(f"Model loaded")
            else:
                pass

                



import paramiko
import time
from aicastle.deepracer.vehicle.api import vehicle_control_module_content

class RebootTrigger:
    def __init__(self, ssh_client, password, reboot_seconds=20):
        self.ssh_client = ssh_client
        self.password = password
        self.rebooted = False
        self.reboot_seconds = reboot_seconds

    def run(self):
        print("Rebooting...")
        stdin, stdout, stderr = ssh_sudo_command(self.ssh_client,  "reboot", self.password)    
        time.sleep(self.reboot_seconds)
        self.rebooted = True

def ssh_connect(hostname, username, password):
    ssh_client = paramiko.SSHClient()
    ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    ssh_client.connect(hostname=hostname, username=username, password=password)

    return ssh_client

def ssh_sudo_command(ssh_client, command, password):
    sudo_command = f"sudo -S -p '' {command}"
    stdin, stdout, stderr = ssh_client.exec_command(sudo_command, get_pty=True)
    time.sleep(1)
    stdin.write(password + "\n")
    stdin.flush()
    return stdin, stdout, stderr

def ssh_remote_content_update(hostname, username, password, remote_file_path, new_content):
    # ssh 연결
    ssh_client = ssh_connect(hostname=hostname, username=username, password=password)
    # 파일 권한 수정
    stdin, stdout, stderr = ssh_sudo_command(ssh_client, f"chmod o+w {remote_file_path}", password)
    # 원격 파일 읽기
    with ssh_client.open_sftp().open(remote_file_path, 'r') as remote_file:
        remote_content = remote_file.read().decode('utf-8')
    
    # 파일 수정할 필요가 없으면 종료
    if remote_content == new_content:
        ssh_client.close()
        reboot_trigger = None
        return reboot_trigger
    # 파일 수정
    else:
        sftp_client = ssh_client.open_sftp()
        with sftp_client.open(remote_file_path, 'w') as remote_file:
            remote_file.write(new_content)

        reboot_trigger = RebootTrigger(ssh_client, password)
        return reboot_trigger

def vehicle_control_module_content_update(hostname, username, password, custom=True):
    remote_file_path="/opt/aws/deepracer/lib/webserver_pkg/lib/python3.8/site-packages/webserver_pkg/vehicle_control.py"
    if custom:
        new_content = vehicle_control_module_content.custom_content
    else:
        new_content = vehicle_control_module_content.origin_content

    reboot_trigger =  ssh_remote_content_update(hostname, username, password, remote_file_path, new_content)
    return reboot_trigger




from setuptools import setup, find_packages
from get_version import read_version

setup(
    name='aicastle',
    version=read_version(),
    packages=find_packages(include=['aicastle', 'aicastle.*']),
    include_package_data=True,
    package_data={
        'aicastle': ['package_data/*'],
    },
    
    # 의존성 필수
    install_requires=[
        'tqdm', 
        "click",
        'pillow',
        'python-dotenv',
        "pyyaml",
    ],

    # 의존성 선택
    extras_require={
        'chat': [
            "streamlit",
            'openai',
            'tiktoken',
            'pathspec',
            'pymupdf',  # fiz
        ],
        'deepracer-vehicle': [
            'tensorflow',
            'opencv-python',
            'requests-toolbelt',
            'beautifulsoup4',
            'lxml',
            'pynput',
            'paramiko',
            # 'ipykernel',
            # 'openai',
            # 'ollama',
        ],
        'deepracer-drfc': [
            "boto3",
            "paramiko",
            'pandas',
            # 'ipykernel',
        ],
    },

    entry_points={
        "console_scripts": [
            "aicastle=aicastle.cli:main",  # aicastle 명령어 실행 엔트리포인트
        ],
    },

    author='aicastle',
    author_email='dev@aicastle.io',
    description='AI Castle Package',
    url='https://github.com/ai-castle/aicastle',
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    zip_safe=False,
)



import os
import json
import numpy as np
import cv2
              
import tensorflow.compat.v1 as tf
tf.compat.v1.disable_v2_behavior()

from tensorflow.compat.v1 import GraphDef, Session, ConfigProto

# 경고 억제
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

class ModelAnalyzer:
    def __init__(
            self, 
            model_folder_path,
            # model_pd_path, 
            # metadata_path
        ):
        self.graph = tf.Graph()
        self.session = Session(
            graph=self.graph, 
            config=ConfigProto(allow_soft_placement=True, log_device_placement=False)
        )
        self.model_folder_path = model_folder_path
        self.metadata_path = os.path.join(model_folder_path, "model_metadata.json")
        self.model_pd_path = os.path.join(model_folder_path, "agent/model.pb")
        # self.metadata_path = metadata_path
        
        self.initialize()


    def initialize(self):
        self._load_metadata()
        with self.graph.as_default():
            self._load_model()             # 세션 생성 + 모델 로드
            self._build_gradcam_ops()      # Grad-CAM용 loss, gradient 텐서 등 미리 구성
            self._get_node_names()         # 모든 노드 이름 저장

    def _load_metadata(self):
        with open(self.metadata_path, "r", encoding="utf-8") as f:
            self.metadata = json.load(f)
            self.training_algorithm = self.metadata['training_algorithm']   # "clipped_ppo" or "sac"
            self.action_space_type = self.metadata['action_space_type'].lower()  # "continuous" or "discrete"
            self.sensor = self.metadata['sensor']  # 예: ["FRONT_FACING_CAMERA"]

    def _load_model(self):
        with tf.io.gfile.GFile(self.model_pd_path, 'rb') as f:
            graph_def = GraphDef()
            graph_def.ParseFromString(f.read())
            tf.import_graph_def(graph_def, name="")
    
    def _build_gradcam_ops(self):
        """
        이산/연속 액션에 따라
         - output_tensor: 최종 policy 텐서
         - loss_tensor  : Grad-CAM 계산용 loss
         - grads_tensor : conv_output에 대한 gradient
        를 정의.
        """
        # 1) Input/Output 텐서
        #   sensor가 ["FRONT_FACING_CAMERA"] 가정
        sensor_name = self.sensor[0]
        if self.training_algorithm == "clipped_ppo":
            self.input_tensor = self.graph.get_tensor_by_name(
                f"main_level/agent/main/online/network_0/{sensor_name}/{sensor_name}:0"
            )
            self.output_tensor = self.graph.get_tensor_by_name(
                 "main_level/agent/main/online/network_1/ppo_head_0/policy:0"
            )
            self.conv_tensor = self.graph.get_tensor_by_name(
                f"main_level/agent/main/online/network_1/{sensor_name}/Conv2d_4/Conv2D:0"
            )
        elif self.training_algorithm == "sac":
            self.input_tensor = self.graph.get_tensor_by_name(
                f"main_level/agent/policy/online/network_0/{sensor_name}/{sensor_name}:0"
            )
            self.output_tensor = self.graph.get_tensor_by_name(
                 "main_level/agent/policy/online/network_0/sac_policy_head_0/policy:0"
            )
            self.conv_tensor = self.graph.get_tensor_by_name(
                f"main_level/agent/policy/online/network_0/{sensor_name}/Conv2d_4/Conv2D:0"
            )

        # 3) 이산 vs. 연속 액션 분기
        if self.action_space_type == "discrete":
            # 이산: argmax를 그래프 내부에서 구하고, one_hot으로 loss 생성
            num_actions = self.output_tensor.shape[-1].value  # 예: N개의 discrete actions
            # 배치=1 가정: self.output_tensor[0] -> shape=(N, )
            self.argmax_idx = tf.argmax(self.output_tensor[0], axis=-1)  # scalar
            one_hot = tf.one_hot(self.argmax_idx, num_actions)           # shape=(N,)
            # loss = Σ (policy[i] * one_hot[i])
            self.loss_tensor = tf.reduce_sum(self.output_tensor[0] * one_hot)
        else:
            # 연속: 예) [steering, speed] -> shape=(1, 2)
            # 전체 합, 또는 특정 차원만 선택 가능
            # 여기서는 전체 합
            self.loss_tensor = tf.reduce_sum(self.output_tensor[0])

        # 4) conv_tensor에 대한 gradient
        self.grads_tensor = tf.gradients(self.loss_tensor, self.conv_tensor)[0]

    def _get_node_names(self):
        """모델 그래프의 모든 노드 이름을 반환."""
        graph_def = self.session.graph.as_graph_def()
        self.node_names = [node.name for node in graph_def.node]

    def bgr2gray(self, img, resize=(160, 120)):
        """BGR -> GRAY, (160,120) 리사이즈."""
        if resize is not None:
            img = cv2.resize(img, resize, interpolation=cv2.INTER_CUBIC)
        return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)

    def predict(self, bgr_img):
        """
        (옵션) 단순 추론용: model.run(self.output_tensor)
        Grad-CAM에는 _build_gradcam_ops()로 처리.
        """
        gray_img = self.bgr2gray(bgr_img)
        inp = np.expand_dims(gray_img, axis=(0, 3))  # shape=(1, 120, 160, 1)
        with self.graph.as_default():
            out = self.session.run(self.output_tensor, feed_dict={self.input_tensor: inp})
        return out[0]  # shape=(action_dim,)

    def visualize_path(self, img_path, alpha=0.5):
        bgr_img = cv2.imread(img_path)
        return self.visualize(bgr_img, alpha=alpha)

    def visualize(self, bgr_img, alpha=0.5):
        """
        Grad-CAM:
        - 한 번의 session.run(...)에서
          1) output_tensor (정책 출력)
          2) (이산이면) argmax_idx
          3) conv_tensor
          4) grads_tensor
        를 모두 구해 사용.
        """
        gray_img = self.bgr2gray(bgr_img)
        inp = np.expand_dims(gray_img, axis=(0, 3))  # (1, 120, 160, 1)

        fetches = [self.output_tensor, self.conv_tensor, self.grads_tensor]
        if self.action_space_type == "discrete":
            fetches.append(self.argmax_idx)

        with self.graph.as_default():
            results = self.session.run(fetches, feed_dict={self.input_tensor: inp})

        # results 순서 해석
        if self.action_space_type == "discrete":
            output_val, conv_val, grads_val, argmax_val = results
        else:
            output_val, conv_val, grads_val = results
            argmax_val = None  # 연속이면 사용X

        # (출력) output_val: shape=(1, action_dim)
        # (conv) conv_val : shape=(1, H, W, Channels)
        # (grads) grads_val: shape=(1, H, W, Channels)

        # Grad-CAM 계산
        weights = np.mean(grads_val, axis=(1, 2))  # (1, Channels)
        cams = np.sum(weights * conv_val, axis=3)  # (1, H, W)
        cam = cams[0]
        cam = cv2.resize(cam, (bgr_img.shape[1], bgr_img.shape[0]))
        cam = np.maximum(cam, 0)

        heatmap = cam / (cam.max() + 1e-5)

        heatmap_color = cv2.applyColorMap(np.uint8(255 * heatmap), cv2.COLORMAP_JET)
        overlay = (alpha * heatmap_color.astype(np.float32)) + ((1 - alpha) * bgr_img.astype(np.float32))
        overlay = overlay / (overlay.max() + 1e-5) * 255.0
        overlay = overlay.astype(np.uint8)

        return overlay, output_val[0], argmax_val

    def check_trainable_vars(self):
        with self.graph.as_default():
            trainables = tf.compat.v1.trainable_variables()
            if len(trainables) == 0:
                print("[INFO] No trainable variables found -> likely a frozen (fully constant) graph.")
            else:
                print(f"[INFO] Found {len(trainables)} trainable variable(s):")
                for v in trainables:
                    print("   -", v.name)

    def close(self):
        if self.session:
            self.session.close()




    # def fine_tune(
    #         self, 
    #         data_folder_path,
    #         train_val_split=0.8,
    #         learning_rate=0.0001,
    #         epochs=10,
    #         patience=3,
    #         batch_size=8,
    #         save_dir="fine_tuned/",
    #         save_period=1,
    #         restore_best_weights=True,
    #     ):

    #     images_path, labels = load_data(data_folder_path)

    #     # 모델 저장 디렉토리 생성
    #     os.makedirs(save_dir, exist_ok=True)
    #     run_id = 0
    #     run_path = os.path.join(save_dir, f"run")
    #     while os.path.exists(run_path):
    #         run_id += 1
    #         run_path = os.path.join(save_dir, f"run_{run_id}")
    #     os.makedirs(run_path)
    #     print(f"Save directory: {run_path}")

    #     # 데이터 분할
    #     num_samples = len(images_path)
    #     train_size = int(train_val_split * num_samples)
    #     indices = np.arange(num_samples)
    #     np.random.shuffle(indices)

    #     train_indices = indices[:train_size]
    #     val_indices = indices[train_size:]

    #     train_images_path = [images_path[i] for i in train_indices]
    #     train_labels = [labels[i] for i in train_indices]
    #     val_images_path = [images_path[i] for i in val_indices]
    #     val_labels = [labels[i] for i in val_indices]

    #     # Dataset 생성
    #     train_dataset = create_tf_dataset(train_images_path, train_labels, batch_size=batch_size)
    #     val_dataset = create_tf_dataset(val_images_path, val_labels, batch_size=batch_size, shuffle=False)

    #     with self.graph.as_default():
    #         # Freeze all layers except the final policy head
    #         for var in self.graph.get_collection(tf.compat.v1.GraphKeys.TRAINABLE_VARIABLES):
    #             if 'policy' not in var.name:
    #                 var.trainable = False
    #             else:
    #                 var.trainable = True

    #         # Define placeholders for inputs and labels
    #         input_placeholder = self.input_tensor  # 이미 정의된 텐서 사용
    #         label_placeholder = tf.compat.v1.placeholder(
    #             tf.float32,
    #             shape=[None] + list(self.output_tensor.shape[1:])  # 배치 크기와 출력 텐서 형태에 맞춤
    #         )

    #         # Define loss and optimizer
    #         if self.action_space_type == "discrete":
    #             loss = tf.reduce_mean(
    #                 tf.compat.v1.nn.softmax_cross_entropy_with_logits(
    #                     logits=self.output_tensor,
    #                     labels=label_placeholder
    #                 )
    #             )
    #         else:
    #             # Continuous: angle과 speed 각각 loss로 계산
    #             loss = tf.reduce_mean(tf.square(self.output_tensor - label_placeholder))

    #         optimizer = tf.compat.v1.train.AdamOptimizer(learning_rate).minimize(loss)

    #         # Initialize variables
    #         self.session.run(tf.global_variables_initializer())

    #         # Create Saver once
    #         saver = tf.train.Saver()

    #         # Track best validation loss and patience
    #         best_val_loss = float('inf')
    #         patience_counter = 0
    #         best_weights = None

    #         # Dataset 반복자 초기화
    #         train_iterator = train_dataset.make_initializable_iterator()
    #         val_iterator = val_dataset.make_initializable_iterator()

    #         # Training loop
    #         for epoch in range(epochs):
    #             # 학습 데이터셋 초기화
    #             self.session.run(train_iterator.initializer)
    #             train_loss = 0
    #             num_batches = 0

    #             # Train on batches
    #             while True:
    #                 try:
    #                     batch_inputs, batch_labels = self.session.run(train_iterator.get_next())
    #                     feed_dict = {
    #                         input_placeholder: batch_inputs,
    #                         label_placeholder: batch_labels
    #                     }
    #                     _, batch_loss = self.session.run([optimizer, loss], feed_dict=feed_dict)
    #                     train_loss += batch_loss
    #                     num_batches += 1
    #                 except tf.errors.OutOfRangeError:
    #                     break

    #             train_loss /= num_batches

    #             # 검증 데이터셋 초기화
    #             self.session.run(val_iterator.initializer)
    #             val_loss = 0
    #             num_val_batches = 0

    #             # Validate on batches
    #             while True:
    #                 try:
    #                     batch_inputs, batch_labels = self.session.run(val_iterator.get_next())
    #                     feed_dict = {
    #                         input_placeholder: batch_inputs,
    #                         label_placeholder: batch_labels
    #                     }
    #                     batch_loss = self.session.run(loss, feed_dict=feed_dict)
    #                     val_loss += batch_loss
    #                     num_val_batches += 1
    #                 except tf.errors.OutOfRangeError:
    #                     break

    #             val_loss /= num_val_batches

    #             print(f"Epoch {epoch + 1}/{epochs}: Train Loss = {train_loss:.4f}, Val Loss = {val_loss:.4f}")

    #             # Save last model
    #             save_last_path = os.path.join(run_path, "last.pd")
    #             saver.save(self.session, save_last_path)

    #             # Save best weights
    #             if val_loss < best_val_loss:
    #                 best_val_loss = val_loss
    #                 patience_counter = 0
    #                 best_weights = self.session.run(self.graph.get_collection(tf.GraphKeys.GLOBAL_VARIABLES))
    #                 save_best_path = os.path.join(run_path, "best_model.pd")
    #                 saver.save(self.session, save_best_path)
    #             else:
    #                 patience_counter += 1

    #             # Early stopping
    #             if patience_counter >= patience:
    #                 print(f"Early stopping at epoch {epoch + 1}. Best validation loss: {best_val_loss:.4f}")
    #                 break

    #             # Save periodically
    #             if save_period > 0 and (epoch + 1) % save_period == 0:
    #                 save_epoch_path = os.path.join(run_path, f"epoch_{epoch + 1}.pd")
    #                 saver.save(self.session, save_epoch_path)

    #         # Restore best weights
    #         if restore_best_weights and best_weights is not None:
    #             for var, best_val in zip(self.graph.get_collection(tf.GraphKeys.GLOBAL_VARIABLES), best_weights):
    #                 self.session.run(var.assign(best_val))
    #             print("Restored best model weights.")
    #             self.model_pd_path = save_best_path
    #         else :
    #             self.model_pd_path = save_last_path

    #     # initialize 메서드 호출
    #     self.initialize()


# class ContinuousConverter:
#     def __init__(self, angle_low=-30.0, angle_high=30.0, speed_low=0.5, speed_high=1.0):
#         """
#         [angle scale]
#             - 'pred': -1 ~ 1,
#             - 'human': -30 ~ 30 (degree)
#             - 'servo': -1 ~ 1
#         [speed scale]
#             - 'pred': -1 ~ 1,
#             - 'human': 0 ~ 4 (m/s)
#             - 'servo': -1 ~ 1
#         """
#         self.angle_low = angle_low
#         self.angle_high = angle_high
#         self.speed_low = speed_low
#         self.speed_high = speed_high
        
#     def pred_to_human(self, angle, speed):
#         if angle is not None:
#             angle = self.angle_low + ((angle + 1.0) / 2.0) * (self.angle_high - self.angle_low)
#         if speed is not None:
#             speed = self.speed_low + ((speed + 1.0) / 2.0) * (self.speed_high - self.speed_low)
#         return angle, speed
    
#     def human_to_pred(self, angle, speed):
#         if angle is not None:
#             angle = 2.0 * (angle - self.angle_low) / (self.angle_high - self.angle_low) - 1.0
#         if speed is not None:
#             speed = 2.0 * (speed - self.speed_low) / (self.speed_high - self.speed_low) - 1.0
#         return angle, speed
    
#     def human_to_servo(self, angle, speed):
#         if angle is not None:
#             angle = angle / 30.0
#         if speed is not None:
#             speed = speed / 4.0
#         return angle, speed
    
#     def servo_to_human(self, angle, speed):
#         if angle is not None:
#             angle = angle * 30.0
#         if speed is not None:
#             speed = speed * 4.0
#         return angle, speed
    
#     def pred_to_servo(self, angle, speed):
#         angle, speed = self.pred_to_human(angle, speed)
#         angle, speed = self.human_to_servo(angle, speed)
#         return angle, speed
    
#     def servo_to_pred(self, angle, speed):
#         angle, speed = self.servo_to_human(angle, speed)
#         angle, speed = self.human_to_pred(angle, speed)
#         return angle, speed



# def load_data(data_folder_path):
#     images_path = []
#     labels = []
#     metadata_path = os.path.join(data_folder_path, "metadata.json")
#     with open(metadata_path, "r") as f:
#         metadata = json.load(f)
#     action_space = metadata['action_space']
#     action_space_type = metadata['action_space_type']
#     if action_space_type == "discrete":
#         for action in action_space:
#             action_folder = os.path.join(data_folder_path, str(action['index']))
#             if not os.path.exists(action_folder):
#                 continue
#             action_images = os.listdir(action_folder)
#             for img_name in action_images:
#                 img_path = os.path.join(action_folder, img_name)
#                 images_path.append(img_path)
#                 labels.append(action['index'])
#     elif action_space_type == "continuous":
#         images_folder = os.path.join(data_folder_path, "images")
#         labels_folder = os.path.join(data_folder_path, "labels")
#         continuous_converter = ContinuousConverter(
#             angle_low=action_space["steering_angle"]["low"],
#             angle_high=action_space["steering_angle"]["high"],
#             speed_low=action_space["speed"]["low"],
#             speed_high=action_space["speed"]["high"]
#         )   
#         for img_name in images_folder:
#             img_path = os.path.join(images_folder, img_name)
#             label_path = os.path.join(labels_folder, img_name)
#             if os.path.exists(label_path):
#                 with open(label_path, "r") as f:
#                     angle, speed = map(float, f.read().strip().split())
#                 angle, speed = continuous_converter.human_to_pred(angle, speed)
#                 images_path.append(img_path)
#                 labels.append([angle, speed])
#     else :
#         raise ValueError("Invalid action_space_type")
#     return images_path, labels


# def tf_preprocess_image(image_path):
#     image = tf.io.read_file(image_path)
#     image = tf.image.decode_jpeg(image, channels=1)  # Grayscale
#     image = tf.image.resize(image, [120, 160])  # Resize to 160x120
#     image = image / 255.0  # Normalize to [0, 1]
#     return image

# def create_tf_dataset(image_paths, labels, batch_size, shuffle=True):
#     dataset = tf.data.Dataset.from_tensor_slices((image_paths, labels))
#     if shuffle:
#         dataset = dataset.shuffle(buffer_size=len(image_paths))
#     dataset = dataset.map(lambda img_path, label: (tf_preprocess_image(img_path), label),
#                           num_parallel_calls=tf.data.AUTOTUNE)
#     dataset = dataset.batch(batch_size).prefetch(buffer_size=tf.data.AUTOTUNE)
#     return dataset




import math
from flask import (Blueprint,
                   jsonify,
                   request)

from deepracer_interfaces_pkg.msg import ServoCtrlMsg
from deepracer_interfaces_pkg.srv import (ActiveStateSrv,
                                          EnableStateSrv,
                                          NavThrottleSrv,
                                          GetCtrlModesSrv)
from webserver_pkg import constants
from webserver_pkg.utility import (api_fail,
                                   call_service_sync)
from webserver_pkg import webserver_publisher_node


VEHICLE_CONTROL_BLUEPRINT = Blueprint("vehicle_control", __name__)


def get_rescaled_manual_speed(categorized_throttle, max_speed_pct):
    return categorized_throttle * max_speed_pct


def get_categorized_manual_throttle(throttle):
    return throttle


def get_categorized_manual_angle(angle):
    return angle


@VEHICLE_CONTROL_BLUEPRINT.route("/api/manual_drive", methods=["PUT", "POST"])
def api_manual_drive():
    """API that publishes control messages to control the angle and throttle in
       manual drive mode.

    Returns:
        dict: Execution status if the API call was successful.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    angle = request.json.get("angle")
    throttle = request.json.get("throttle")
    max_speed = request.json.get("max_speed")

    if angle is None:
        return api_fail("angle is required")
    if throttle is None:
        return api_fail("throttle is required")
    if max_speed is None:
        return api_fail("max_speed is required")

    if angle < -1.0 or angle > 1.0:
        return api_fail("angle out of range")
    if throttle < -1.0 or throttle > 1.0:
        return api_fail("throttle out of range")

    webserver_node.get_logger().info(f"Angle: {angle}  Throttle: {throttle}")

    # Create the servo message.
    msg = ServoCtrlMsg()
    # bound the throttle value based on the categories defined
    msg.angle = -1.0 * get_categorized_manual_angle(angle)
    categorized_throttle = get_categorized_manual_throttle(throttle)
    msg.throttle = -1.0 * get_rescaled_manual_speed(categorized_throttle, max_speed)
    webserver_node.pub_manual_drive.publish(msg)
    return jsonify({"success": True})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/drive_mode", methods=["PUT", "POST"])
def api_set_drive_mode():
    """API to toggle the drive mode between Autonomous/Manual mode.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    drive_mode = request.json.get("drive_mode")
    if drive_mode is None:
        return jsonify({"success": False, "reason": "drive_mode must be set."})

    webserver_node.get_logger().info(f"Changed the vehicle state to {drive_mode}")
    drive_mode_state = 0 if drive_mode == "manual" else 1

    try:
        vehicle_state_req = ActiveStateSrv.Request()
        vehicle_state_req.state = drive_mode_state
        vehicle_state_res = call_service_sync(webserver_node.vehicle_state_cli,
                                              vehicle_state_req)
        if vehicle_state_res and (vehicle_state_res.error == 0):
            return jsonify(success=True)
        else:
            webserver_node.get_logger().error("Vehicle state service call failed")
            return jsonify(success=False, reason="Error")

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach vehicle state server: {ex}")
        return jsonify({"success": False,
                        "reason": "Unable to reach vehicle state server."})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/start_stop", methods=["PUT", "POST"])
def api_set_start_stop():
    """API to call the enable_state service to start and stop the vehicle.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    start_stop = request.json.get("start_stop")
    if start_stop is None:
        return jsonify({"success": False, "reason": "start_stop must be set."})

    webserver_node.get_logger().info(f"Changed the enable state to {start_stop}")
    start_stop_state = False if start_stop == "stop" else True
    try:
        enable_state_req = EnableStateSrv.Request()
        enable_state_req.is_active = start_stop_state
        enable_state_res = call_service_sync(webserver_node.enable_state_cli, enable_state_req)
        if enable_state_res and (enable_state_res.error == 0):
            return jsonify({"success": True})
        else:
            return jsonify({"success": False, "reason": "Error"})

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach enable state server: {ex}")
        return jsonify({"success": False,
                        "reason": "Unable to reach enable state server."})


@VEHICLE_CONTROL_BLUEPRINT.route("/api/max_nav_throttle", methods=["PUT", "POST"])
def max_nav_throttle():
    """API to call the navigation_throttle service to set the throttle scale in the
       autonomous mode.

    Returns:
        dict: Execution status if the API call was successful and error
              reason if failed.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    nav_throttle = request.json.get("throttle")
    if nav_throttle is None:
        return jsonify({"success": False, "reason": "value must be set."})
    webserver_node.get_logger().info(f"Setting max navigation throttle to {nav_throttle}")
    try:
        set_throttle_req = NavThrottleSrv.Request()
        set_throttle_req.throttle = nav_throttle / constants.MAX_AUTO_THROTTLE_VAL
        set_throttle_res = call_service_sync(webserver_node.set_throttle_cli, set_throttle_req)
        if set_throttle_res and (set_throttle_res.error == 0):
            return jsonify({"success": True})
        else:
            return jsonify(success=False, reason="Failed to call the navigation throttle service")
    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach navigation throttle server: {ex}")
        return jsonify(success=False, reason="Unable to reach navigation throttle server")


@VEHICLE_CONTROL_BLUEPRINT.route("/api/control_modes_available", methods=["GET"])
def control_modes_available():
    """API to call the GetCtrlModesCountSrv service to get the list of available modes
       in ctrl_pkg (autonomous/manual/calibration).

    Returns:
        dict: Execution status if the API call was successful, list of available modes
              and error reason if call fails.
    """
    webserver_node = webserver_publisher_node.get_webserver_node()
    webserver_node.get_logger().info("Providing the number of available modes")
    try:
        get_ctrl_modes_req = GetCtrlModesSrv.Request()
        get_ctrl_modes_res = call_service_sync(webserver_node.get_ctrl_modes_cli,
                                               get_ctrl_modes_req)

        control_modes_available = list()
        for mode in get_ctrl_modes_res.modes:
            control_modes_available.append(constants.MODE_DICT[mode])

        data = {
            "control_modes_available": control_modes_available,
            "success": True
        }
        return jsonify(data)

    except Exception as ex:
        webserver_node.get_logger().error(f"Unable to reach get ctrl modes service: {ex}")
        return jsonify(success=False, reason="Error")
